"""
A python client to parse the schema from any graphql endpoint.
It allows you to do the following:

- access the schema of any graphql endpoint
  - allowed queries/mutations/subscriptions using
      API.schema.queries, API.schema.mutations, API.schema.subscriptions
  - allowed fields and args for queries/mutations/subscriptions
      API.schema.fetch(<query/mutation/subscription name>)
  - explore fields of type defs
      API.schema.explore(<typedef>)

- query the graphql endpoint with format
    API.query(<query type>, args=<args>).fetch(<list of fields>)

- murate the graphql endpoint with format
    API.mutate(<mutation type>, args=<args>).fetch(<list of fields>)
"""

from graphqlclient import GraphQLClient
import json
from epigraph import schema
from epigraph import utilities


class GraphQLAPI(object):
    """
    Class to parse the schema from any graphql endpoint.
    It allows you to do the following:
    - access the schema of any graphql endpoint
      - allowed queries/mutations/subscriptions
      - allowed fields and args for queries/mutations/subscriptions
      - explore fields of type defs
    - query the graphql endpoint with format
        API.query(<query type>, args=<args>).fetch(<list of fields>)
    - murate the graphql endpoint with format
        API.mutate(<mutation type>, args=<args>).fetch(<list of fields>)

    Parameters
    ==========
    url : str
        url of the graphql endpoint.
    """

    def __init__(self, url):
        self.client = GraphQLClient(url)
        schema_quert_str = """{
          __schema {
            queryType {
              name
              interfaces {
                name
              }
            }
            mutationType{
              name
              interfaces{
                name
              }
            }
            subscriptionType{
              name
              interfaces{
                name
              }
            }
            types {
              name
              kind
              ofType {
                kind
                name
              }
              enumValues {
                name
              }
              fields {
                name
                type{
                  name
                  kind
                  ofType{
                    name
                    kind
                    ofType{
                      name
                      kind
                      ofType{
                        name
                        kind
                      }
                    }
                  }
                }
                args {
                  name
                  defaultValue
                  type {
                    name
                    kind
                    ofType {
                      name
                      kind
                      ofType{
                        name
                        kind
                        ofType{
                          name
                          kind
                        }
                      }
                    }
                  }
                }
              }
              inputFields {
                name
                defaultValue
                type {
                  kind
                  name
                  ofType {
                    kind
                    name
                    ofType{
                      name
                      kind
                      ofType{
                        name
                        kind
                      }
                    }
                  }
                }
              }
            }
          }
        }"""
        schema_json = json.loads(self.client.execute(schema_quert_str))
        self.schema = schema.Schema(schema_json)

    def query(self, query_type, args={}):
        """
        Method to start the query string.
        This method needs to be appended with `.args()` method. 
        Parameters
        ==========

        query_type: str
            the name of the query you want to make.
        """
        # query_type = inputs.get('query_type')
        assert (
            query_type in self.schema.queries
        ), "Wrong query_type. Choices are: {}".format(self.schema.queries)
        self.query_mutation = query_type
        self.query_str = "{" + str(query_type) + "}"
        return self

    def mutate(self, mutation_type):
        """
        Method to start the mutation string.
        This method needs to be appended with `.args()` method. 
        Parameters
        ==========

        mutation_type: str
            the name of the mutation you want to make.
        """
        assert (
            mutation_type in self.schema.mutations
        ), "Wrong mutation_type. Choices are: {}".format(self.schema.mutations)
        self.query_mutation = mutation_type
        self.query_str = "mutation{" + str(mutation_type) + "}"
        return self

    def args(self, args={}):
        """
        This method adds the arguments to the query string.
        This method needs to be appended with `.fetch()` method. 
        Paramaters
        ==========
        args: dict
            the input args if the query requires them.
        """
        if args:
            assert isinstance(
                args, dict
            ), "Wrong data structure for args: Must be a dictionary"
        if args == {}:
            assert (
                len(list(self.schema.fetch(self.query_mutation)[0].keys())) == 0
            ), "Missing arguments. Choices are: {}".format(
                list(self.schema.fetch(self.query_mutation)[0].keys())
            )
        for a in args:
            assert (
                a in self.schema.fetch(self.query_mutation)[0].keys()
            ), "Wrong input argument key '{}'. Choices are: {}".format(
                a, list(self.schema.fetch(self.query_mutation)[0].keys())
            )
        if self.query_str.endswith("}"):
            self.query_str = self.query_str[:-1]
        if len(args) > 0:
            if self.query_mutation in self.schema.mutations: 
                query_str = utilities.dict_to_arg_str(args)
                self.query_str += query_str
            else:
                arg_str = utilities.dict_to_arg_str(args)
                arg_str = arg_str.replace("{", "")
                arg_str = arg_str.replace(",{", "")
                arg_str = arg_str.replace("[", "{")
                arg_str = arg_str.replace("]", "")
                self.query_str += arg_str
        self.query_str += "}"
        return self

    def fetch(self, return_struct=None):
        """
        Method to allow the user to complete the query/mutation and execute it at the graphql endpoint.
        This method needs to be called after `.query() or .mutate()` method.
        eg. API.query().args().fetch() or API.mutate().args().fetch().

        Parameters
        ==========
        return_struct: list
            list of all the fields you would like to request from the graphql endpoint.
            If `None`, the code tries to figure it out from the schema and returns all fields.
        
        Returns
        =======
        data: dict
            Returned data structure of a query/mutation.
        """
        if self.query_str.endswith("}"):
            self.query_str = self.query_str[:-1]
        if not return_struct:
            return_str = str(self.schema.fetch(self.query_mutation)[1])
            return_str = return_str.replace("[", "")
            return_str = return_str.replace("]", "")
            return_str = return_str.replace(": ", "")
            return_str = return_str.replace(":", "")
            for d in self.schema.default_types:
                return_str = return_str.replace("'{}'".format(d), "")
            return_str = return_str.replace("'", "")
        else:
            return_str = utilities.dict_to_query_str(return_struct)
            return_str = return_str.replace(":", "")
            return_str = return_str.replace("{", "")
            return_str = return_str.replace(",{", "")
            return_str = return_str.replace("[", "{")
            return_str = return_str.replace("]", "")
            while return_str.count("{") != return_str.count("}"):
                if return_str.count("{") > return_str.count("}"):
                    return_str += "}"
                else:
                    return_str.insert(0, "{")

        self.query_str += return_str
        self.query_str += "}"
        data = json.loads(self.client.execute(self.query_str))
        return data



import copy


class BaseType(object):
    def __init__(self, kind, name):
        self.kind = kind
        self.name = name


class Types(object):
    def __init__(self, kind, name, ofType):
        self.kind = kind
        self.name = name
        self.ofType = None
        if ofType:
            self.ofType = BaseType(ofType["kind"], ofType["name"])


class Field(object):
    def __init__(self, name, field_type, args):
        self.name = name
        self.type = Types(field_type["kind"], field_type["name"], field_type["ofType"])
        self.args = []
        for a in args:
            self.args.append(Args(a["name"], a["defaultValue"], a["type"]))


class Args(object):
    def __init__(self, name, default_value, arg_type):
        self.name = name
        self.default_value = default_value
        self.type = Types(arg_type["kind"], arg_type["name"], arg_type["ofType"])


class RootTypes(object):
    def __init__(self, kind, name, ofType, fields):
        self.kind = kind
        self.name = name
        self.ofType = None
        if ofType:
            self.ofType = BaseType(ofType["kind"], ofType["name"])
        self.fields = []
        if fields:
            for f in fields:
                self.fields.append(Field(f["name"], f["type"], f["args"]))


class Schema(object):
    """
    Parses the schema of the graphql endpoint.
    returns all queries, mutations and subscriptions allowed by the graphql endpoint.
    Allows you to fetch any type and see it's underlying fields.

    Parameters
    ==========
    schema_json: parsed json data
        schema from the graphwl endpoint.
    """

    def __init__(self, schema_json):
        self.default_types = [
            "String",
            "Int",
            "Float",
            "Boolean",
            "DateTime",
            "ID",
            "Enum",
            "__Schema",
            "__Type",
            "__TypeKind",
            "__Field",
            "__InputValue",
            "__EnumValue",
            "__Directive",
            "__DirectiveLocation",
        ]
        self.type_map = {
            "INPUT_OBJECT": [dict],
            "OBJECT": [dict],
            "SCALAR": [int, float, bool, str],
        }
        self.__parse_schema(schema_json)

    def __parse_schema(self, schema):
        """
        Parses the schema and extracts query, mutation and subscription types.
        Parameters
        ==========
        schema: json
            loaded json schema from the graphql endpoint.
        """
        self.query_type = schema["data"]["__schema"]["queryType"]
        self.query_interfaces = self.query_type["interfaces"]
        self.mutation_type = schema["data"]["__schema"]["mutationType"]
        self.subscription_type = schema["data"]["__schema"]["subscriptionType"]
        self.all_types = schema["data"]["__schema"]["types"]
        queries = []
        mutations = []
        subscriptions = []
        for s in self.all_types:
            if self.query_type and s["name"] == self.query_type["name"]:
                queries.append(
                    RootTypes(s["kind"], s["name"], s["ofType"], s["fields"])
                )
            if self.mutation_type and s["name"] == self.mutation_type["name"]:
                mutations.append(
                    RootTypes(s["kind"], s["name"], s["ofType"], s["fields"])
                )
            if self.subscription_type and s["name"] == self.subscription_type["name"]:
                subscriptions.append(
                    RootTypes(s["kind"], s["name"], s["ofType"], s["fields"])
                )
        self.queries = []
        for r in queries:
            for f in r.fields:
                self.queries.append(f.name)
        self.mutations = []
        for r in mutations:
            for f in r.fields:
                self.mutations.append(f.name)
        self.subscriptions = []
        for r in subscriptions:
            for f in r.fields:
                self.subscriptions.append(f.name)

    def fetch(self, fetch_type):
        """
        Fetches the arguments and return types for a query, mutation, subscription type.

        Parameters
        ==========
        fetch_type: str
        
        Returns
        =======
            tuple of (arguments, fields).
        """
        assert (
            fetch_type in self.queries + self.mutations + self.subscriptions
        ), "Wrong input type. Choices are: \nQueries: {}\nMutations: {}\n Subscriptions:{}\n".format(
            self.queries, self.mutations, self.subscriptions
        )
        if fetch_type in self.queries:
            base_type = self.query_type["name"]
            return self.__fetch(base_type, fetch_type)
        if fetch_type in self.mutations:
            base_type = self.mutation_type["name"]
            return self.__fetch(base_type, fetch_type)
        if fetch_type in self.subscriptions:
            base_type = self.subscription_type["name"]
            return self.__fetch(base_type, fetch_type)

    def explore(self, fetch_type):
        """
        Explores the fields of any typedef.

        Parameters
        ==========
        fetch_type: str
            typedef you want to explore.
        
        Returns
        =======
        dict of fields for the typedef.
        """
        if fetch_type in self.queries + self.mutations + self.subscriptions:
            return self.fetch(fetch_type)

        return self.__explore_struct(fetch_type)

    def __explore_struct(self, base_type):
        """
        Helper method to explore the typedef.

        Parameters
        ==========
        base_type: str
            typedef you want to explore.
        
        Returns
        =======
        dict of fields for the typedef.
        """
        return_type = {}
        for t in self.all_types:
            if t["name"] == base_type:
                if t["fields"]:
                    for f in t["fields"]:
                        name_to_check = f["type"]["name"]
                        check = f["type"]
                        islist = False
                        if check["kind"] in ["LIST"]:
                            islist = True
                        while name_to_check is None:
                            if check["kind"] in ["LIST"]:
                                islist = True
                            name_to_check = check["ofType"]["name"]
                            check = check["ofType"]
                        if islist:
                            name_to_check = [name_to_check]
                        return_type[f["name"]] = name_to_check
                if t["inputFields"]:
                    for f in t["inputFields"]:
                        name_to_check = f["type"]["name"]
                        check = f["type"]
                        islist = False
                        if check["kind"] in ["LIST"]:
                            islist = True
                        while name_to_check is None:
                            if check["kind"] in ["LIST"]:
                                islist = True
                            name_to_check = check["ofType"]["name"]
                            check = check["ofType"]
                        if islist:
                            name_to_check = [name_to_check]
                        return_type[f["name"]] = name_to_check
                if t["enumValues"]:
                    # struct[c] = 'Enum'
                    enums = []
                    for f in t["enumValues"]:
                        enums.append(f["name"])
                    return_type = enums
        return return_type

    def __fetch(self, base_type, fetch_type):
        """
        Fetches the arguments and fields of requested query/mutation/schema.
        
        Parameters
        ==========
        base_type: str
            name of the base query/mutation/schema.
        fetch_type: str
            query/mutation/scubscription that needs to be fetch.

        Returns
        =======
        args: dict
            Arguments that the fetch type needs
            eg.
            query{
              Planet(args){
                climate
              }
            }
        return_type: dict
            Return types that the fetch type needs.
            eg.
            query{
              Planet(name: "Earth"){
                return_type
              }
            }

        """
        return_type = None
        args = None
        for t in self.all_types:
            if t["name"] == base_type:
                for f in t["fields"]:
                    if fetch_type == f["name"]:
                        if f["type"]["kind"] in ["LIST", "NON_NULL"]:
                            return_type = [
                                self.__fetch_return_type(f["type"]["ofType"])
                            ]
                        else:
                            return_type = self.__fetch_return_type(
                                {"kind": f["type"]["kind"], "name": f["type"]["name"]}
                            )
                        args = self.__fetch_args(f["args"])

        return args, return_type

    def __fetch_args(self, args):
        """
        Helper method to fetch the args of a query/mutation/subscription type.
        
        Parameters
        ==========
        args: dict

        Returns
        =======
        args: dict
            nicely formated
            {name: type}
        """
        if args == []:
            return
        outer_struct = {}
        for a in args:
            if a["type"]["name"] in self.default_types:
                outer_struct[a["name"]] = a["type"]["name"]
                continue
            struct = {}
            name_to_check = a["type"]["name"]
            check = a["type"]
            while name_to_check is None:
                name_to_check = check["ofType"]["name"]
                check = check["ofType"]
            to_check = [check["name"]]
            return_type = check["name"]
            check = to_check
            checked = []
            while len(check):
                for c in to_check:
                    checked.append(c)
                    if c in self.default_types:
                        struct[return_type] = c
                        check.remove(c)
                        check = list(set(check))
                        continue
                    for t in self.all_types:
                        if t["name"] == c:
                            dstruct = {}
                            if t["fields"]:
                                for f in t["fields"]:
                                    further, value = self.__explore_further(f)
                                    if further:
                                        if isinstance(value, list):
                                            if value[0] not in checked:
                                                check.append(value[0])
                                        else:
                                            if value not in checked:
                                                check.append(value)
                                        dstruct[f["name"]] = value
                                    else:
                                        dstruct[f["name"]] = value
                                    struct[c] = dstruct
                            if t["inputFields"]:
                                for f in t["inputFields"]:
                                    further, value = self.__explore_further(f)
                                    if further:
                                        if isinstance(value, list):
                                            if value[0] not in checked:
                                                check.append(value[0])
                                        else:
                                            if value not in checked:
                                                check.append(value)
                                        dstruct[f["name"]] = value
                                    else:
                                        dstruct[f["name"]] = value
                                    struct[c] = dstruct
                            if t["enumValues"]:
                                # struct[c] = 'Enum'
                                struct[c] = []
                                for f in t["enumValues"]:
                                    struct[c].append(f["name"])
                    if c in check:
                        check.remove(c)
                    check = list(set(check))
                to_check = check
            reduced_struct = copy.deepcopy(struct)
            while len(reduced_struct) > 0:
                to_check = list(reduced_struct.keys())
                for c in to_check:
                    if isinstance(struct[c], dict):
                        for k in struct[c].keys():
                            if isinstance(struct[c][k], list):
                                if struct[c][k][0] in struct:
                                    struct[c][k][0] = struct[struct[c][k][0]]
                                    # break
                            else:
                                if struct[c][k] in struct:
                                    struct[c][k] = struct[struct[c][k]]
                                    # break
                    del reduced_struct[c]
            outer_struct[a["name"]] = struct[return_type]
        return outer_struct

    def __fetch_return_type(self, return_type):
        """
        Helper method to fetch the fields of a query/mutation/subscription type.
        
        Parameters
        ==========
        return_type: dict

        Returns
        =======
        return_type: dict
            nicely formated
            {name: type}
        """
        if return_type["name"] in self.default_types:
            return return_type["name"]
        struct = {}
        name_to_check = return_type["name"]
        check = return_type
        while name_to_check is None:
            name_to_check = check["ofType"]["name"]
            check = check["ofType"]
        to_check = [check["name"]]
        return_type = check["name"]
        check = to_check
        checked = []
        while len(check):
            for c in to_check:
                checked.append(c)
                if c in self.default_types:
                    struct[return_type] = c
                    check.remove(c)
                    check = list(set(check))
                    continue
                for t in self.all_types:
                    if t["name"] == c:
                        dstruct = {}
                        if t["fields"]:
                            for f in t["fields"]:
                                further, value = self.__explore_further(f)
                                if further:
                                    if isinstance(value, list):
                                        if value[0] not in checked:
                                            check.append(value[0])
                                    else:
                                        if value not in checked:
                                            check.append(value)
                                    dstruct[f["name"]] = value
                                else:
                                    dstruct[f["name"]] = value
                                struct[c] = dstruct
                        if t["enumValues"]:
                            struct[c] = "Enum"
                if c in check:
                    check.remove(c)
                check = list(set(check))
            to_check = check
        reduced_struct = copy.deepcopy(struct)
        while len(reduced_struct) > 0:
            to_check = list(reduced_struct.keys())
            for c in to_check:
                if isinstance(struct[c], dict):
                    for k in struct[c].keys():
                        if isinstance(struct[c][k], list):
                            if struct[c][k][0] in struct:
                                struct[c][k][0] = struct[struct[c][k][0]]
                        else:
                            if struct[c][k] in struct:
                                struct[c][k] = struct[struct[c][k]]
                del reduced_struct[c]
        return struct[return_type]

    def __explore_further(self, field):
        """
        Helper method to determine if the field needs to be explored further
        in case it is a nested field with internal data structures.

        Parameters
        ==========
        field: dict
            check if this field needs to be explored

        Return
        ======
        bool, name : whether the field needs to be explored further and the name of the field.
        """
        name_to_check = field["type"]["name"]
        check = field["type"]
        return_type = None
        if check["kind"] == "LIST":
            return_type = "LIST"
        elif check["kind"] == "ENUM":
            return_type = "ENUM"
        while name_to_check is None:
            name_to_check = check["ofType"]["name"]
            if check["kind"] == "LIST":
                return_type = "LIST"
            elif check["kind"] == "ENUM":
                return_type = "ENUM"
            check = check["ofType"]
        if return_type == "LIST":
            if check["name"] not in self.default_types:
                return True, [check["name"]]
            else:
                return False, [check["name"]]
        elif return_type == "ENUM":
            return False, "Enum"
        elif field["type"]["kind"] in ["INPUT_OBJECT", "OBJECT"]:
            return True, check["name"]
        else:
            return False, check["name"]



def replace_str(data):
    """
    Function to recursively traverse a json data structure and
    replace strings with double quotes.

    Parameters
    ==========
    data: json data structure

    Returns
    =======
    data: updated json
    """
    if isinstance(data, dict):
        return {k: replace_str(v) for k, v in data.items()}
    elif isinstance(data, list):
        return [replace_str(i) for i in data]
    else:
        return '"{}"'.format(data) if isinstance(data, str) else data


def dict_to_query_str(dict_object):
    """
    Converts a dictionary object to Graphql string.
    """
    input_str = str(dict_object)
    input_str = input_str.replace("'", "")
    while input_str.count("{") != input_str.count("}"):
        if input_str.count("{") > input_str.count("}"):
            input_str += "}"
        else:
            input_str.insert(0, "{")
    return input_str


def dict_to_arg_str(args):
    """
    Converts an arguments dictionary object to Graphql argument string.
    """
    args = replace_str(args)
    input_str = "("
    for i, k in enumerate(args):
        if i > 0:
            input_str += ","
        if isinstance(args[k], dict):
            sub_str = dict_to_query_str(args[k])
            input_str += str(k) + ": " + sub_str
        else:
            if isinstance(args[k], str):
                input_str += str(k) + ":" + args[k]
            else:
                input_str += str(k) + ":" + str(args[k])
    input_str += ")"
    return input_str



from epigraph import graphql_api

API = graphql_api.GraphQLAPI(url="https://swapi.graph.cool/")


def test_schema():
    response = API.schema.queries
    assert "Planet" in response
    response = API.schema.mutations
    assert "createPlanet" in response
    response = API.schema.subscriptions
    assert "Planet" in response
    response = API.schema.fetch("Planet")
    assert "name" in response[0]
    assert "climate" in response[1]
    response = API.schema.fetch("FilmcharactersPerson")
    assert "birthYear" in response
    assert "eyeColor" in response
    assert "homeworld" in response


def test_query():
    args = {"name": "Alderaan"}
    payload = ["climate"]
    response = API.query("Planet").args=(args).fetch(payload)
    assert response == {"data": {"Planet": {"climate": ["temperate"]}}}


def test_mutations():
    args = {"name": "Earth"}
    payload = ["name"]
    response = API.mutate("createPlanet").args=(args).fetch(payload)
    assert response["data"] is not None
    if response["data"]:
        assert "createPlanet" in response["data"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['epigraph']

package_data = \
{'': ['*']}

install_requires = \
['graphqlclient>=0.2,<0.3']

setup_kwargs = {
    'name': 'epigraph',
    'version': '2020.7.27.1',
    'description': 'Python API for graphql',
    'long_description': '# CHIME/FRB API\n\n|  **`Release`**  |   **`Style`**   |   **`Documentation`**   |\n|-----------------|-----------------|-------------------------|\n| [![PyPI version](https://img.shields.io/pypi/v/epigraph.svg)](https://pypi.org/project/epigraph/) | [![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://black.readthedocs.io/en/stable/)|[![Documentation: sphinx](https://img.shields.io/badge/documentation-docs-brightgreen)](https://chimefrb.github.io/epigraph/build/html/index.html)|\n\n--------\n\n# epigraph\nAutomated GraphQL API\n\n### Installation\n```\npip install epigraph\n```\n\n### Running the API\n```python\nIn [1]: from epigraph import graphql_api\n\nIn [2]: API = graphql_api.GraphQLAPI(url="https://swapi.graph.cool/")\n\nIn [3]: API.query(\'Planet\').args({\'name\': "Alderaan"}).fetch([\'climate\'])\nOut[3]: {"data": {"Planet": {"climate": ["temperate"]}}}\n\n```\n',
    'author': 'Shiny Brar',
    'author_email': 'charanjotbrar@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/CHIMEFRB/epigraph',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
